=begin
= NArrayMiss
 NArrayMiss is a additional class to NArray
 To use NArrayMiss class, you need invoking "require 'narray_miss'"
 in your script.
=end

require 'narray'


class NArrayMiss

=begin
== Class Constants
--- NArrayMiss::BYTE
--- NArrayMiss::SINT
--- NArrayMiss::INT
--- NArrayMiss::SFLOAT
--- NArrayMiss::FLOAT
--- NArrayMiss::SCOMPLEX
--- NArrayMiss::COMPLEX
--- NArrayMiss::OBJECT
=end

  BYTE = NArray::BYTE
  SINT = NArray::SINT
  INT = NArray::INT
  SFLOAT = NArray::SFLOAT
  FLOAT = NArray::FLOAT
  SCOMPLEX = NArray::SCOMPLEX
  COMPLEX = NArray::COMPLEX
  OBJECT = NArray::OBJECT

  class << self
    alias :__new__ :new
    private :__new__
  end

  def initialize(array, mask)
    if array.shape!=mask.shape
      raise "array and mask must have the same shape"
    end
    @array = array
    @mask = mask
  end
  private :initialize

=begin
== Class Methods
--- NArrayMiss.new(typecode, size, ...)
--- NArrayMiss.byte(size, ...)
--- NArrayMiss.sint(size, ...)
--- NArrayMiss.int(size, ...)
--- NArrayMiss.sfloat(size, ...)
--- NArrayMiss.float(size, ...)
--- NArrayMiss.scomplex(size, ...)
--- NArrayMiss.complex(size, ...)
--- NArrayMiss.object(size, ...)
--- NArrayMiss[](indices, ...)
--- NArrayMiss.to_nam(array)
--- NArrayMiss.to_nam_no_dup(array)
=end

  def self.new(*arg)
    array = NArray.new(*arg)
    mask = NArray.byte(*arg[1..-1])
    __new__(array, mask)
  end
  def self.byte(*arg)
    NArrayMiss.new(BYTE,*arg)
  end
  def self.sint(*arg)
    NArrayMiss.new(SINT,*arg)
  end
  def self.int(*arg)
    NArrayMiss.new(INT,*arg)
  end
  def self.sfloat(*arg)
    NArrayMiss.new(SFLOAT,*arg)
  end
  def self.float(*arg)
    NArrayMiss.new(FLOAT,*arg)
  end
  def self.scomplex(*arg)
    NArrayMiss.new(SCOMPLEX,*arg)
  end
  def self.complex(*arg)
    NArrayMiss.new(COMPLEX,*arg)
  end
  def self.object(*arg)
    NArrayMiss.new(OBJECT,*arg)
  end
  def self.[](*arg)
    NArrayMiss.to_nam(*arg)
  end
  def self.to_nam_no_dup(*arg)
    if arg.length > 2 || arg.length==0 then
      raise("NArrayMiss.to_nar( array [,mask]] )")
    end

    array = arg[0]
    if array.type < Numeric then array = NArray[array] end
    if array.type == Array then array = NArray.to_na(array) end
    if array.type != NArray then
      raise("array must be NArray or Array")
    end

    if arg.length==2 then
      mask = arg[1]
      if mask.type < Numeric then mask = [mask] end
      if mask.type == TrueClass then mask = [1] end
      if mask.type == Array then
	tmp = NArray.byte(mask.length)
	tmp[]=mask
	mask = tmp
      end
      if mask.type == FalseClass then
	mask = NArray.byte(*array.shape)
      end
      if mask.type == TrueClass then
	mask = NArray.byte(*array.shape).fill(1)
      end
      if !(mask.type==NArray && mask.typecode==BYTE) then
	  raise("mask must be Numeric, Array, true, false or NArray(byte)")
      end
    else
      mask = NArray.byte(*array.shape).fill(1)
    end
    __new__(array,mask)
  end
  def self.to_nam(*arg)
    if arg[0].type!=TrueClass && arg[0].type!=FalseClass then
      arg[0] = arg[0].dup
    end
    if arg.length==2 && arg[1].type!=TrueClass && arg[1].type!=FalseClass then
      arg[1] = arg[1].dup
    end
    NArrayMiss.to_nam_no_dup(*arg)
  end


=begin
== Class Instance Methods
=end


=begin
=== NArrayMiss information
--- NArraMiss#dim
--- NArraMiss#rank
--- NArraMiss#shape
--- NArraMiss#total
--- NArraMiss#rank_total
--- NArraMiss#typecode
=end

  def dim
    @array.dim
  end
  def rank
    @array.rank
  end
  def shape
    @array.shape
  end
  def total
    @array.total
  end

  def rank_total(*arg)
    @array.rank_total(*arg)
  end

  def typecode
    @array.typecode
  end


=begin
=== Slicing Array
--- NArraMiss#[](index)
--- NArraMiss#slice(index)
--- NArraMiss#[]=(index, value)
=end

  def [](*arg)
    NArrayMiss.to_nam_no_dup(@array[*arg],@mask[*arg])
  end
  def slice(*arg)
    NArrayMiss.to_nam_no_dup(@array.slice(*arg),@mask.slice(*arg))
  end

  def []=(*arg)
    if arg.length==1 then
      if arg[0].type == NArrayMiss then
	@array[] = arg[0].get_array
	@mask[] = arg[0].get_mask
      else
	@array[] = arg[0]
      end
    else
      if arg[-1].type == NArrayMiss then
	@array[*arg[0..-2]] = arg[-1].get_array
	@mask[*arg[0..-2]] = arg[-1].get_mask
      else
	@array[*arg[0..-2]] = arg[-1]
      end
    end
  end


=begin
=== Filling values
--- NArraMiss#indgen([start[,step]])
--- NArraMiss#indgen!([start[,step]])
--- NArraMiss#fill(value)
--- NArraMiss#fill!(value)
--- NArraMiss#random(max)
--- NArraMiss#random!(max)
--- NArraMiss#randomn(max)
=end

  for operator in ["indgen","fill","random"]
    routine = <<-EOL
    def #{operator}(*arg)
      obj = self.dup
      obj.#{operator}!(*arg)
      obj
    end
    def #{operator}!(*arg)
      @array[@mask] = @array[@mask].#{operator}!(*arg)
      self
    end
    EOL
    eval(routine)
  end
  def randomn
    obj = self.dup
    obj[@mask] = @array[@mask].rondomn
    obj
  end


=begin
=== Arithmetic operator
--- NArraMiss#-@
--- NArraMiss#+(other)
--- NArraMiss#-(other)
--- NArraMiss#*(other)
--- NArraMiss#/(other)
--- NArraMiss#%(other)
--- NArraMiss#**(other)
--- NArraMiss#abs
--- NArraMiss#add!(other)
--- NArraMiss#sbt!(other)
--- NArraMiss#mul!(other)
--- NArraMiss#div!(other)
--- NArraMiss#mod!(other)
--- NArraMiss#mul_add(other, dim, ...)
=end

  def -@
    @array[@mask] = -@array[@mask]
    self
  end
  for operator in ["+","-","*","/","%","**"]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1 #{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end
  def abs
    @array[@mask] = @array[@mask].abs
    self
  end

  for operator in ["add!","sbt!","mul!","div!","mod!"]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1.#{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end
  def mul_add(*arg)
    if arg.length==1 then
      (self*arg[0]).sum
    else
      (self*arg[0]).sum(*arg[1..-1])
    end
  end


=begin
=== Bitwise operator (only for byte, sint and int)
--- NArraMiss#~@
--- NArraMiss#&(other)
--- NArraMiss#|(other)
--- NArraMiss#^(other)
=end

  def ~@
    @array = ~@array
    self
  end
  for operator in ["&","|","^"]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1 #{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end


=begin
=== Comparison
--- NArraMiss#eq(other)
--- NArraMiss#ne(other)
--- NArraMiss#gt(other)
--- NArraMiss#ge(other)
--- NArraMiss#lt(other)
--- NArraMiss#le(other)
--- NArraMiss#>(other)
--- NArraMiss#>=(other)
--- NArraMiss#<(other)
--- NArraMiss#<=(other)
--- NArraMiss#and(other)
--- NArraMiss#or(other)
--- NArraMiss#xor(other)
--- NArraMiss#not(other)
=end

  for operator in ["eq","ne","gt","ge","lt","le"]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1.#{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end
  for operator in [">",">=","<","<="]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1 #{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end

  for operator in ["and","or","xor","not"]
    routine = <<-EOL
    def #{operator}(arg)
      term1,term2,mask,flag = routine1(arg)
      result = term1.#{operator} term2
      routine2(result,mask,flag)
    end
    EOL
    eval(routine)
  end

#  def ==(arg)
#    if art.kind_of?(NArrayMiss) then
#      @array==arg.get_array && @mask==arg.get_mask
#    else
#      false
#    end
#  end


=begin
=== Statistics
--- NArraMiss#accum(dim, ...)
--- NArraMiss#sum(dim, ...)
--- NArraMiss#min(dim, ...)
--- NArraMiss#max(dim, ...)
--- NArraMiss#mean(dim, ...)
--- NArraMiss#stddev(dim, ...)
=end

  def accum(*arg)
    if @mask.count_true == 0 then
      return nil
    else
      array = @array.dup
      array[@mask.where2[1]] = 0
      NArrayMiss.to_nam_no_dup(array.accum(*arg),
			       @mask.to_type(NArray::INT).accum(*arg).ne(0))
    end
  end

  for operator in ["sum","min","max"]
    str = {"sum"=>"0","min"=>"array.max","max"=>"array.min"}[operator]
    routine = <<-EOL
    def #{operator}(*arg)
      if @mask.count_true == 0 then
	return nil
      else
	array = @array.dup
	array[@mask.where2[1]] = #{str}
	if arg.length==0 then
	  array.#{operator}
	else
	  NArrayMiss.to_nam_no_dup(array.#{operator}(*arg),
				   @mask.to_type(NArray::INT).sum(*arg).ne(0))
	end
      end
    end
    EOL
    eval(routine)
  end
  def mean(*arg)
    self.sum(*arg)/@mask.to_type(NArray::INT).sum(*arg)
  end
  def stddev(*arg)
    count = @mask.to_type(NArray::INT).sum(*arg)
    mean = self.mean(*arg)
    var = ((self**2).sum(*arg) +count*mean**2 -self.sum(*arg)*mean*2)/
      NArrayMiss.to_nam(count-1,count>1)
    NMMath::sqrt(var)
  end
#  def median(*arg)
#  end


=begin
=== Sort
--- NArraMiss#sort(dim)
--- NArraMiss#sort_index(dim)
=end

  for operator in ["sort","sort_index"]
    routine = <<-EOL
    def #{operator}(*arg)
      obj=NArrayMiss.new(@array.typecode,*@array.shape)
      if arg.length==0 then
	obj[@mask] = @array[@mask].#{operator}
	obj
      else
	shape=@array.shape
	shape[*arg]=1
	nshape = NArray.to_na(shape)
	slice = shape[*nshape.ne(1).where]
	nslice = NArray.to_na(slice)
	index = NArray.object(@mask.rank)
	index[*nshape.eq(1).where] = true
	obj = NArrayMiss.new(@array.typecode,*@array.shape)
	for i in 0..nslice.sum-1
	  index[*nshape.ne(1).where] = pos(i,nslice)
	  mask = NArray.byte(*@array.shape).fill(0)
	  mask[*index.to_a] = 1
	  mask = @mask&mask
	  if mask.count_true != 0 then
	    obj[mask] = @array[mask].#{operator}
	  end
	end
	obj
      end
    end
    EOL
    eval(routine)
  end


=begin
=== Transpose
--- NArraMiss#transpose(dim0, dim1, ...)
=end

  def transpose(*arg)
    obj = self.dup
    obj[] = @array.transpose(*arg)
    obj.set_mask(@mask.transpose(*arg))
    obj
  end


=begin
=== Changing Shapes of indices
--- NArraMiss#reshape(size, ...)
--- NArraMiss#reshape!(size, ...)
--- NArraMiss#shape=(size, ...)
--- NArraMiss#newdim(size, ...)
--- NArraMiss#newdim!(size, ...)
--- NArraMiss#rewrank!(size, ...)
--- NArraMiss#rewrank=(size, ...)
=end

  def reshape!(*arg)
    @array = @array.reshape!(*arg)
    @mask = @mask.reshape!(*arg)
    self
  end
  def reshape(*arg)
    obj = self.dup
    obj.reshape!(*arg)
  end
  alias :shape= :reshape!
  def newdim!(*arg)
    @array = @array.newdim!(*arg)
    @mask = @mask.newdim!(*arg)
    self
  end
  alias :rewrank! :newdim!
  alias :rewrank= :newdim!
  def newdim(*arg)
    obj = self.dup
    obj.newdim!(*arg)
  end
  alias :rewrank :newdim


=begin
=== Type conversion
--- NArraMiss#floor
--- NArraMiss#ceil
--- NArraMiss#round
--- NArraMiss#to_i
--- NArraMiss#to_f
--- NArraMiss#to_a
--- NArraMiss#to_na
--- NArraMiss#to_s
--- NArraMiss#to_string
=end

  for operator in ["floor","ceil","round","to_i","to_f"]
    routine = <<-EOL
    def #{operator}
      NArrayMiss.to_nam_no_dup(@array.#{operator}, @mask)
    end
    EOL
    eval(routine)
  end
  def to_a
    @array.to_a
  end
  def to_na(*arg)
    @array
  end
  def to_s
    @array.to_s
  end
  def to_string
    @array.to_string
  end


=begin
=== Iteration
--- NArraMiss#each{|i| ...}
--- NArraMiss#collect{|i| ...}
--- NArraMiss#collect!{|i| ...}
=end

  def each
    for i in 0..self.total-1
      yield(self[i])
    end
  end
  def collect
    obj = NArrayMiss.new(@array.typecode,*@array.shape)
    for i in 0..self.total-1
      obj[i] = yield(self[i])
    end
  end
  def collect!
    for i in 0..self.total-1
      self[i] = yield(self[i])
    end
  end


=begin
=== Boolean / mask related
--- NArraMiss#count_false
--- NArraMiss#count_true
--- NArraMiss#mask(mask)
--- NArraMiss#all?
--- NArraMiss#any?
--- NArraMiss#none?
--- NArraMiss#where
--- NArraMiss#where2
=end

  def count_false
    if @array.typecode==BYTE then
      @array.count_false-@mask.count_false
    else
      raise("cannot count_true NArrayMiss except BYTE type")
    end
  end
  def count_true
    if @array.typecode==BYTE then
      (@array&@mask).count_true
    else
      raise("cannot count_true NArrayMiss except BYTE type")
    end
  end
  def mask(arg)
    if arg.type==NArrayMiss then
      arg = arg.get_array&arg.get_mask
    end
    self.set_mask(@mask&arg)
  end

  def all?
    @array[@mask].all?
  end
  def any?
    @array[@mask].any?
  end
  def none?
    @array[@mask].none?
  end

  def where
    (@array&@mask).where
  end
  def where2
    self.where-@mask.where
  end




=begin
=== Complex compound number
--- NArraMiss#real
--- NArraMiss#imag
--- NArraMiss#conj
--- NArraMiss#angle
--- NArraMiss#imag=(other)
--- NArraMiss#im
=end

  def real
    NArrayMiss.to_nam_no_dup(@array.real,@mask)
  end
  def imag
    NArrayMiss.to_nam_no_dup(@array.imag,@mask)
  end
  def conj
    NArrayMiss.to_nam_no_dup(@array.conj,@mask)
  end
  def angle
    NArrayMiss.to_nam_no_dup(@array.angle,@mask)
  end
  def imag=(arg)
    @array.image=(arg)
    self
  end
  def im
    NArrayMiss.to_nam_no_dup(@array.im,@mask)
  end


=begin
=== Byte swap
--- NArraMiss#swap_byte
--- NArraMiss#hton
--- NArraMiss#ntoh
--- NArraMiss#htov
--- NArraMiss#vtoh
=end

  def swap_byte
    obj = self.dup
    obj[] = @array.swap_byte
    obj
  end
  def hton
    NArray.to_nam(@array.hton,@mask.hton)
  end
  alias :ntoh :hton
  def htov
    NArray.to_nam(@array.htov,@mask.htov)
  end
  alias :vtoh :htov


=begin
=== Mask and missing_value
--- NArraMiss#set_missing(index)
--- NArraMiss#unset_missing(index)
--- NArraMiss#set_mask(mask)
--- NArraMiss#set_missing_value(value)
--- NArraMiss#get_mask
--- NArraMiss#get_array
--- NArraMiss#valid?
--- NArraMiss#count_valid
--- NArraMiss#count_invalid
=end

  def set_missing(pos)
    @mask[pos] = 0
  end
  def unset_missing(pos)
    @mask[pos] = 1
  end
  def set_mask(mask)
    if @array.shape != mask.shape then
      raise("mask.shape must be same as array")
    end
    if mask.type == Array then
      tmp = NArray.byte(*mask.shape)
      tmp[true] = mask
      mask = tmp
    end
    if mask.type == NArray then
      if mask.typecode != BYTE then
	raise("mask must be NArray.byte or Array")
      end
      @mask = mask
    else
      raise("mask must be NArray.byte or Array")
    end
  end

  def set_missing_value!(val)
    @array[@mask.where2[1]] = val
    self
  end
  def set_missing_value(val)
    obj = NArrayMiss.new(@array.typecode,@array.shape)
    obj.set_mask(@mask)
    obj[@mask] = @array[@mask]
    obj.set_missing_value!(val)
  end

  def get_mask
    @mask
  end
  def get_array
    @array
  end

  def valid?
    where = self.get_mask.where2
    tf = Array.new(self.total)
    for i in where[0]
      tf[i] = true
    end
    for i in where[1]
      tf[i] = false
    end
    tf
  end

  def count_valid(*arg)
    if arg.length==0 then
      @mask.count_true
    else
      @mask.to_type(NArray::INT).sum(*arg)
    end    
  end
  def count_invalid(*arg)
    if arg.length==0 then
      @mask.count_false
    else
      NArray.int(*@mask.shape).fill(1).sum(*arg)-@mask.to_type(NArray::INT).sum(*arg)
    end
  end


=begin
=== others
--- NArraMiss#integer?
--- NArraMiss#complex?
--- NArraMiss#dup
--- NArraMiss#coerce(object)
--- NArraMiss#inspect
=end

  def integer?
    @array.integer?
  end
  def complex?
    @array.complex?
  end


  def dup
    NArrayMiss.to_nam(@array,@mask)
  end

  def coerce(x)
    if x.type<Numeric then
      [NArrayMiss.new(NArray[x].typecode,*self.shape).fill(x),self]
    elsif x.type==Array || x.type==NArray then
      [NArrayMiss.to_nam(x), self]
    else
      raise("donnot know how to cange #{x.type} to NArrayMiss")
    end
  end


  def inspect
    "array -> " + @array.inspect + "\nmask  -> " + @mask.inspect
  end

#  private
  private
  def pos(n,shape)
    rank = shape.length
    result = NArray.int(rank)
    m=n
    for i in 0..rank-2
      j = rank-1-i
      result[j] = m/shape[j-1]
      m = m%shape[j-1]
    end
    result[0] = m
    result
  end
  def routine1(arg)
    mask = @mask
    flag=true
    if arg.type < Numeric then
      term1 = @array.mask(mask)
      term2 = arg
    elsif arg.type == Array then
      term1 = @array.mask(mask)
      term2 = NArray.to_na(arg).mask(mask)
    elsif arg.type == NArray then
      term1 = @array.mask(mask)
      term2 = arg.mask(mask)
    elsif arg.type == NArrayMiss then
      mask = mask & arg.get_mask
      term1 = @array.mask(mask)
      term2 = arg.get_array.mask(mask)
    else
      term1,term2 = arg.coerce(self)
      flag=false
    end
    [term1,term2,mask,flag]
  end
  def routine2(result,mask,flag)
    if flag then
      obj = NArrayMiss.new(result.typecode,*@array.shape)
      obj.set_mask(mask)
      obj[mask] = result
      obj
    else
      result
    end
  end
    

end



module NMMath

  for operator in ["sqrt","exp","log","log10","log2",
                   "atan2","sin","cos","tan","sinh","cosh","tanh",
                   "asin","acos","atan","asinh","acosh","atanh"]
    routine = <<-EOL
    def #{operator}(x)
      if x.type < Numeric || x.type==Array || x.type==NArray then
	NMath::#{operator}(x)
      elsif x.type == NArrayMiss then
	obj = NArrayMiss.new(x.typecode,*x.shape)
	mask = x.get_mask
	obj.set_mask(mask)
	obj[mask] = NMath::#{operator}(x.get_array[mask])
	obj[mask.where2[1]] = x[mask.where2[1]]
        obj
      end
    end
    module_function :#{operator}
    EOL
    eval(routine)
  end

end

